import json
import sys

import yt.wrapper as yt

from tests.utils import (
    GENERIC_DOCKER_IMAGE,
    TractoCli,
    get_data_path,
    make_cli_args,
    make_run_config,
    run_config_file,
)
from tests.yt_instances import YtInstance
from tractorun.backend.generic import GenericBackend
from tractorun.cli.tractorun_runner import make_configuration
from tractorun.mesh import Mesh
from tractorun.operation_log import OperationLogMode
from tractorun.private.process_manager import OutputType
from tractorun.run import run
from tractorun.sidecar import (
    RestartPolicy,
    Sidecar,
)
from tractorun.stderr_reader import StderrMode
from tractorun.toolbox import Toolbox


def test_configuration() -> None:
    _, _, config = make_configuration(make_cli_args())
    assert config.operation_log_mode == OperationLogMode.default

    _, _, config = make_configuration(make_cli_args("--operation-log-mode", OperationLogMode.realtime_yt_table.value))
    assert config.operation_log_mode == OperationLogMode.realtime_yt_table

    run_config = make_run_config({"operation_log_mode": OperationLogMode.realtime_yt_table.value})
    with run_config_file(run_config) as run_config_path:
        _, _, config = make_configuration(["--run-config-path", run_config_path])
    assert config.operation_log_mode == OperationLogMode.realtime_yt_table

    with run_config_file(run_config) as run_config_path:
        _, _, config = make_configuration(
            ["--run-config-path", run_config_path, "--operation-log-mode", OperationLogMode.default.value]
        )
    assert config.operation_log_mode == OperationLogMode.default


def validate_logs(yt_client: yt.YtClient, incarnation: int, yt_path: str) -> None:
    for index in range(4):
        messages: list[tuple[str, str]] = []
        for raw in yt_client.read_table(f"{yt_path}/logs/{incarnation}/workers/{index}"):
            assert raw["datetime"] is not None
            messages.append(
                (raw["message"], raw["fd"]),
            )
        assert (f"first stdout line {index} {incarnation}", OutputType.stdout) in messages
        assert (f"first stderr line {index} {incarnation}", OutputType.stderr) in messages
        assert (f"second stdout line {index} {incarnation}", OutputType.stdout) in messages
        assert (f"second stderr line {index} {incarnation}", OutputType.stderr) in messages

    for index in range(4):
        raws = [raw for raw in yt_client.read_table(f"{yt_path}/logs/{incarnation}/sidecars/{index}")]
        assert len(raws) == 1
        raw = raws[0]
        assert raw["datetime"] is not None
        output_type = OutputType.stdout if index % 2 == 0 else OutputType.stderr
        assert raw["fd"] == output_type
        assert raw["message"] == f"sidecar_writes_something_{incarnation}_{output_type}"


def test_pickle(yt_instance: YtInstance, yt_path: str) -> None:
    yt_client = yt_instance.get_client()

    def checker(toolbox: Toolbox) -> None:
        self_index = toolbox.coordinator.get_self_index()
        inc = toolbox.coordinator.get_incarnation_id()
        print(f"first stdout line {self_index} {inc}", file=sys.stdout)
        print(f"first stderr line {self_index} {inc}", file=sys.stderr)
        print(f"second stdout line {self_index} {inc}", file=sys.stdout)
        print(f"second stderr line {self_index} {inc}", file=sys.stderr)

    mesh = Mesh(node_count=2, process_per_node=2, gpu_per_process=0)
    for incarnation in range(2):
        sidecar_stdout_message = f"sidecar_writes_something_{incarnation}_{OutputType.stdout}"
        sidecar_stderr_message = f"sidecar_writes_something_{incarnation}_{OutputType.stderr}"
        run(
            checker,
            backend=GenericBackend(),
            yt_path=yt_path,
            operation_log_mode=OperationLogMode.realtime_yt_table,
            sidecars=[
                Sidecar(
                    command=["python3", "-c", f'import sys; print("{sidecar_stdout_message}", file=sys.stdout)'],
                    restart_policy=RestartPolicy.ON_FAILURE,
                ),
                Sidecar(
                    command=["python3", "-c", f'import sys; print("{sidecar_stderr_message}", file=sys.stderr)'],
                    restart_policy=RestartPolicy.ON_FAILURE,
                ),
            ],
            mesh=mesh,
            yt_client=yt_client,
            docker_image=GENERIC_DOCKER_IMAGE,
        )

        validate_logs(yt_client=yt_client, incarnation=incarnation, yt_path=yt_path)


def test_cli(yt_instance: YtInstance, yt_path: str) -> None:
    yt_client = yt_instance.get_client()

    for incarnation in range(2):
        sidecar_stdout_message = f"sidecar_writes_something_{incarnation}_{OutputType.stdout}"
        sidecar_stderr_message = f"sidecar_writes_something_{incarnation}_{OutputType.stderr}"
        tracto_cli = TractoCli(
            command=["python3", "/tractorun_tests/log_table_script.py"],
            args=[
                "--yt-path",
                yt_path,
                "--mesh.node-count",
                "2",
                "--mesh.process-per-node",
                "2",
                "--operation-log-mode",
                OperationLogMode.realtime_yt_table.value,
                "--sidecar",
                json.dumps(
                    {
                        "command": ["python3", "-c", f'import sys; print("{sidecar_stdout_message}", file=sys.stdout)'],
                        "restart_policy": RestartPolicy.ON_FAILURE,
                    },
                ),
                "--sidecar",
                json.dumps(
                    {
                        "command": ["python3", "-c", f'import sys; print("{sidecar_stderr_message}", file=sys.stderr)'],
                        "restart_policy": RestartPolicy.ON_FAILURE,
                    },
                ),
                "--bind-local",
                f"{get_data_path('../data/log_table_script.py')}:/tractorun_tests/log_table_script.py",
                "--proxy-stderr-mode",
                StderrMode.primary,
            ],
        )
        op_run = tracto_cli.run()
        assert op_run.is_exitcode_valid()
        assert op_run.is_operation_state_valid(yt_client=yt_client, job_count=2)

        validate_logs(yt_client=yt_client, incarnation=incarnation, yt_path=yt_path)



import abc
import base64
import datetime
from multiprocessing import (
    Process,
    Queue,
)
import pickle
import queue
import sys
import time

import attrs
from yt import type_info
from yt import wrapper as yt

from tractorun.private.sidecar import (
    SidecarIndex,
    SidecarRun,
)
from tractorun.private.training_dir import TrainingDir
from tractorun.private.worker import (
    WorkerIndex,
    WorkerRun,
)


BULK_TABLE_WRITE_SIZE = 100
WAIT_LOG_RECORDS_TIMEOUT = 5
IO_QUEUE_MAXSIZE = 10000
YT_LOG_WRITER_JOIN_TIMEOUT = 10
QUEUE_TIMEOUT = 0.01


class _NoMessages:
    pass


class _LastMessage:
    pass


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class LogRecord:
    _message: str
    _datetime: datetime.datetime
    _fd: str

    def _datetime_to_unixtime(self, datetime_obj: datetime.datetime) -> int:
        # workaround for https://github.com/ytsaurus/ytsaurus/issues/309
        return int(time.mktime(datetime_obj.timetuple()))

    @staticmethod
    def get_yt_schema() -> yt.schema.TableSchema:
        schema = yt.schema.TableSchema()
        schema.add_column("datetime", type_info.Datetime)
        schema.add_column("message", type_info.String)
        schema.add_column("fd", type_info.String)
        return schema

    def to_dict(self) -> dict:
        return {
            "message": self._message,
            "fd": self._fd,
            "datetime": self._datetime_to_unixtime(self._datetime),
        }


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class LogHandler(abc.ABC):
    @abc.abstractmethod
    def process(self, record: LogRecord) -> None:
        pass

    @abc.abstractmethod
    def stop(self) -> None:
        pass


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class LogHandlerFactory(abc.ABC):
    @abc.abstractmethod
    def create_for_worker(self, worker_index: WorkerIndex, worker_run: WorkerRun) -> LogHandler:
        pass

    @abc.abstractmethod
    def create_for_sidecar(self, sidecar_index: SidecarIndex, sidecar_run: SidecarRun) -> LogHandler:
        pass


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class YTLogHandler(LogHandler):
    _queue: Queue
    _log_writer: Process

    @classmethod
    def start(cls, yt_client_config: str, log_path: str) -> "YTLogHandler":
        io_queue: Queue = Queue(maxsize=IO_QUEUE_MAXSIZE)
        yt_log_writer = Process(
            target=YtLogWriter(
                yt_client_config=yt_client_config,
                queue=io_queue,
                log_path=log_path,
            )
        )
        yt_log_writer.start()
        return YTLogHandler(queue=io_queue, log_writer=yt_log_writer)

    def process(self, record: LogRecord) -> None:
        try:
            self._queue.put(record, timeout=QUEUE_TIMEOUT)
        except queue.Full:
            print("io queue is full, can't write data to the table", file=sys.stderr)

    def stop(self) -> None:
        self._queue.put(_LastMessage())
        self._log_writer.join(timeout=YT_LOG_WRITER_JOIN_TIMEOUT)
        if self._log_writer.is_alive():
            self._log_writer.kill()
            self._log_writer.join()
        self._log_writer.close()
        self._queue.close()


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class YTLogHandlerFactory(LogHandlerFactory):
    _yt_client_config: str
    _training_dir: TrainingDir

    def create_for_worker(self, worker_index: WorkerIndex, worker_run: WorkerRun) -> YTLogHandler:
        log_path = f"{self._training_dir.worker_logs_path}/{worker_index}"
        return YTLogHandler.start(log_path=log_path, yt_client_config=self._yt_client_config)

    def create_for_sidecar(self, sidecar_index: SidecarIndex, sidecar_run: SidecarRun) -> YTLogHandler:
        log_path = f"{self._training_dir.sidecar_logs_path}/{sidecar_index}"
        return YTLogHandler.start(log_path=log_path, yt_client_config=self._yt_client_config)


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class YtLogWriter:
    _yt_client_config: str
    _queue: Queue
    _log_path: str

    def __call__(self) -> None:
        return self._write_log()

    def _write_log(self) -> None:
        yt_config = pickle.loads(base64.b64decode(self._yt_client_config))
        yt_client = yt.YtClient(config=yt_config)
        yt_client.create(
            "table",
            self._log_path,
            attributes={"schema": LogRecord.get_yt_schema().to_yson_type()},
        )
        got_last_message = False
        while not got_last_message:
            messages: list[dict[str, str | int]] = []
            while True:
                message = self._get_next_message()
                match message:
                    case _NoMessages():
                        time.sleep(WAIT_LOG_RECORDS_TIMEOUT)
                    case _LastMessage():
                        got_last_message = True
                        break
                    case LogRecord():
                        messages.append(message.to_dict())
                        if len(messages) >= BULK_TABLE_WRITE_SIZE:
                            break
            if not messages and not got_last_message:
                time.sleep(WAIT_LOG_RECORDS_TIMEOUT)
                continue
            yt_client.write_table(
                yt.TablePath(self._log_path, append=True),
                messages,
            )

    def _get_next_message(self) -> LogRecord | _NoMessages:
        try:
            return self._queue.get(timeout=QUEUE_TIMEOUT)
        except queue.Empty:
            return _NoMessages()



import sys
import time
import typing

import numpy as np
import orbax.checkpoint as ocp

# ytpath patches other libs
# we have to keep this import
import ytpath  # noqa

# try to use jax
from tractorun.backend.generic import GenericBackend
from tractorun.run import prepare_and_get_toolbox


def timed_run(action: typing.Callable) -> None:

    start = time.time()
    action()
    print(f"Time: {time.time() - start:.2f}s", file=sys.stderr)


def gen_tree() -> list:
    rand_tensor = np.random.rand(25 * 1024**2).astype(np.float32)
    return [{"t" + str(j): rand_tensor for j in range(4)} for _ in range(4)]


def main() -> None:
    toolbox = prepare_and_get_toolbox(backend=GenericBackend())

    user_config = toolbox.get_user_config()
    use_ocdbt = user_config["use_ocdbt"]
    use_zarr3 = user_config["use_zarr3"]

    checkpoint_path = f"yt://{user_config['checkpoint_path']}/tensorproxy"

    checkpointer = ocp.Checkpointer(ocp.PyTreeCheckpointHandler(use_ocdbt=use_ocdbt, use_zarr3=use_zarr3))
    print("Generating", file=sys.stderr)
    my_tree = gen_tree()
    print("Saving")
    timed_run(lambda: checkpointer.save(checkpoint_path, my_tree))

    checkpointer = ocp.Checkpointer(ocp.PyTreeCheckpointHandler(use_ocdbt=use_ocdbt, use_zarr3=use_zarr3))
    print("Restoring", file=sys.stderr)
    timed_run(lambda: checkpointer.restore(checkpoint_path))


if __name__ == "__main__":
    main()



from collections import Counter
import json
from pathlib import (
    Path,
    PosixPath,
)
import shutil
from typing import Iterable
import warnings
import zipfile

import attrs
import yt.wrapper as yt
from yt.wrapper.errors import YtResolveError

from tractorun.bind import (
    BindCypress,
    BindLocal,
)
from tractorun.exception import TractorunConfigurationError


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class PackedBind:
    yt_path: str
    local_path: str


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class BindsPacker:
    _files: Iterable[BindLocal]
    _dirs: Iterable[BindLocal]

    @classmethod
    def from_binds(cls, binds: Iterable[BindLocal]) -> "BindsPacker":
        files: list[BindLocal] = []
        dirs: list[BindLocal] = []

        for bind in sorted(binds, key=lambda x: x.source):
            if Path(bind.source).is_dir():
                dirs.append(bind)
            else:
                files.append(bind)
        return BindsPacker(files=files, dirs=dirs)

    def _pack_file(self, archive_path: str, archive_name: str, source_path: str, destination_name: str) -> PackedBind:
        with zipfile.ZipFile(archive_path + ".zip", "w") as zipf:
            zipf.write(source_path, arcname=destination_name)
        return PackedBind(
            yt_path=f"{archive_name}.zip",
            local_path=f"{archive_path}.zip",
        )

    def _pack_dir(self, archive_path: str, archive_name: str, source_path: str) -> PackedBind:
        shutil.make_archive(archive_path, "zip", source_path)
        return PackedBind(
            yt_path=f"{archive_name}.zip",
            local_path=f"{archive_path}.zip",
        )

    def pack(self, base_result_path: str | Path) -> list[PackedBind]:
        base_result_path = Path(base_result_path)
        result: list[PackedBind] = []

        for idx, bind in enumerate(self._dirs):
            archive_name = f"__dir_{idx}"
            archive_path = base_result_path / archive_name
            packed_bind = self._pack_dir(
                archive_path=str(archive_path),
                archive_name=archive_name,
                source_path=bind.source,
            )
            result.append(packed_bind)

        for idx, bind in enumerate(self._files):
            archive_name = f"__file_{idx}"
            archive_path = base_result_path / archive_name
            destination_name = Path(bind.destination).name
            packed_bind = self._pack_file(
                archive_path=str(archive_path),
                archive_name=archive_name,
                source_path=bind.source,
                destination_name=destination_name,
            )
            result.append(packed_bind)

        return result

    def unpack(self) -> None:
        # Runs inside a job
        for idx, bind in enumerate(self._dirs):
            path_w = f"__dir_{idx}.zip"
            with zipfile.ZipFile(path_w, "r") as zipf:
                zipf.extractall(path=bind.destination)
        for idx, bind in enumerate(self._files):
            path_w = f"__file_{idx}.zip"
            destination = str(Path(bind.destination).parent)
            with zipfile.ZipFile(path_w, "r") as zipf:
                zipf.extractall(path=destination)

    def to_env(self) -> str:
        return json.dumps(
            {
                "files": [attrs.asdict(bind) for bind in self._files],  # type: ignore
                "dirs": [attrs.asdict(bind) for bind in self._dirs],  # type: ignore
            },
        )

    @classmethod
    def from_env(cls, content: str) -> "BindsPacker":
        parsed = json.loads(content)
        files = [BindLocal(source=record["source"], destination=record["destination"]) for record in parsed["files"]]
        dirs = [BindLocal(source=record["source"], destination=record["destination"]) for record in parsed["dirs"]]
        return BindsPacker(files=files, dirs=dirs)


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class PackedLib:
    archive_name: str
    path: str


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class BindsLibPacker:
    _paths: Iterable[str]

    def pack(self, base_result_path: str) -> list[PackedLib]:
        paths = [Path(path).resolve() for path in self._paths]
        lib_to_original_path = {path.name: path for path in paths}
        non_uniq_names = [name for name, count in Counter(lib_to_original_path.keys()).items() if count > 1]
        assert len(non_uniq_names) == 0, f"Some libraries have the same names: {non_uniq_names}"

        result: list[PackedLib] = []
        for lib_name, original_path in lib_to_original_path.items():
            archive_path = Path(base_result_path) / f"__{lib_name}"
            root_dir, base_dir = original_path.resolve().parent, original_path.name
            shutil.make_archive(str(archive_path), "zip", str(root_dir), base_dir)
            result.append(
                PackedLib(
                    archive_name=lib_name,
                    path=f"{archive_path}.zip",
                )
            )

        return result


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class _YtNode:
    name: str
    path: str


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class BindsCypressProcessor:
    _yt_client: yt.YtClient
    _binds: Iterable[BindCypress]

    def process(self) -> list[BindCypress]:
        new_binds: list[BindCypress] = []
        for bind in self._binds:
            node_type = self._get_source_node_type(bind.source)
            if node_type == "map_node":
                for node in self._get_files(bind.source):
                    new_binds.append(
                        BindCypress(
                            source=node.path,
                            destination=f"{bind.destination}/{node.name}",
                            attributes=bind.attributes,
                        )
                    )
            else:
                new_binds.append(bind)
        return new_binds

    def _get_files(self, path: str, prefix: str = "") -> Iterable[_YtNode]:
        nodes: list[_YtNode] = []
        for node_name in self._yt_client.list(path):
            node_path = f"{path}/{node_name}"
            node_type = self._get_source_node_type(node_path)
            match node_type:
                case "map_node":
                    nested_nodes = self._get_files(
                        node_path,
                        prefix=str(PosixPath(prefix) / node_name),
                    )
                    nodes.extend(nested_nodes)
                case "file":
                    nodes.append(
                        _YtNode(
                            name=str(PosixPath(prefix) / node_name),
                            path=node_path,
                        ),
                    )
                case _:
                    warnings.warn(f"Skip {node_path} because it is not a file, but {node_type}")
        return nodes

    def _get_source_node_type(self, path: str) -> str:
        try:
            return str(self._yt_client.get_attribute(path, "type"))
        except YtResolveError as e:
            raise TractorunConfigurationError(f"Source path for path {path} doesn't exist") from e



import os
from typing import Generator
import warnings

import pytest
import yt_yson_bindings

from tests.utils import (
    get_data_path,
    get_random_string,
)
from tests.yt_instances import (
    YtInstance,
    YtInstanceExternal,
    YtInstanceTestContainers,
)
from tractorun.private.yt_cluster import TractorunClusterConfig


@pytest.fixture(scope="session")
def yt_instance() -> Generator[YtInstance, None, None]:
    yt_mode = os.environ.get("YT_MODE", "testcontainers")
    if yt_mode == "testcontainers":
        with YtInstanceTestContainers() as yt_instance:
            yield yt_instance
    elif yt_mode == "external":
        proxy_url = os.environ["YT_PROXY"]
        yt_token = os.environ.get("YT_TOKEN")
        assert yt_token is not None
        yield YtInstanceExternal(proxy_url=proxy_url, token=yt_token)
    else:
        raise ValueError(f"Unknown yt_mode: {yt_mode}")


@pytest.fixture(scope="session")
def yt_base_dir(yt_instance: YtInstance) -> str:
    yt_client = yt_instance.get_client()

    path = f"//tmp/tractorun_tests/run_{get_random_string(4)}"
    yt_client.create("map_node", path, recursive=True)
    return path


@pytest.fixture(scope="function")
def yt_path(yt_instance: YtInstance, yt_base_dir: str) -> str:
    yt_client = yt_instance.get_client()
    path = f"{yt_base_dir}/{get_random_string(4)}"
    yt_client.create("map_node", path)
    return path


@pytest.fixture(scope="function")
def cluster_config() -> TractorunClusterConfig:
    return TractorunClusterConfig(
        cypress_link_template="https://yt.tracto.ai/yt/navigation?path={path}",
        job_stderr_link_template="https://yt.tracto.ai/api/yt/yt/api/v3/get_job_stderr?operation_id={operation_id}&job_id={job_id}&dump_error_into_response=true",
    )


@pytest.fixture
def cluster_config_path(cluster_config: TractorunClusterConfig, yt_instance: YtInstance, yt_path: str) -> str:
    config_path = f"{yt_path}/tractorun_config.yaml"
    yt_client = yt_instance.get_client()
    yt_client.set(config_path, cluster_config.to_dict())
    return config_path


@pytest.fixture(scope="session")
def mnist_ds_path(yt_base_dir: str, yt_instance: YtInstance) -> Generator[str, None, None]:
    table_path = f"{yt_base_dir}/{get_random_string(4)}"

    yt_client = yt_instance.get_client()
    yt_client.create(
        "table",
        table_path,
        attributes={  # TODO: type_v3?
            "schema": [
                {"name": "data", "type": "string"},
                {"name": "labels", "type": "string"},
            ]
        },
    )

    with open(get_data_path("mnist_small.yson"), "rb") as mnist_file:
        parsed_data = yt_yson_bindings.load(mnist_file, yson_type="list_fragment")
        yt_client.write_table(
            table=table_path,
            input_stream=parsed_data,
        )

    yield table_path

    yt_client.remove(table_path)


@pytest.fixture(scope="session")
def can_test_jax() -> bool:
    # we can't import jax inside orbax container
    # it's a session fixture because we can't import jax twice (I don't know why)
    try:
        import jax  # noqa: F401
    except RuntimeError as e:
        if "This version of jaxlib was built using AVX instructions" in str(e):
            warnings.warn(str(e))
            return False
    return True



import json

from tests.utils import (
    TENSORPROXY_DOCKER_IMAGE,
    TractoCli,
    get_data_path,
    run_config_file,
)
from tests.yt_instances import YtInstance


def test_run_script(yt_instance: YtInstance, yt_path: str) -> None:
    yt_client = yt_instance.get_client()

    tracto_cli = TractoCli(
        command=["python3", "/tractorun_tests/tensorproxy_script.py"],
        docker_image=TENSORPROXY_DOCKER_IMAGE,
        args=[
            "--mesh.gpu-per-process",
            "0",
            "--tensorproxy.enabled",
            "1",
            "--resources.memory-limit",
            "428000000000",
            "--yt-path",
            yt_path,
            "--user-config",
            json.dumps({"use_ocdbt": False, "use_zarr3": False, "checkpoint_path": yt_path}),
            "--bind-local",
            f"{get_data_path('../data/tensorproxy_script.py')}:/tractorun_tests/tensorproxy_script.py",
        ],
    )
    op_run = tracto_cli.run()
    assert op_run.is_exitcode_valid()
    assert op_run.is_operation_state_valid(yt_client=yt_client, job_count=1)


def test_run_script_with_config(yt_instance: YtInstance, yt_path: str) -> None:
    yt_client = yt_instance.get_client()

    run_config = {
        "mesh": {
            "node_count": 1,
            "process_per_node": 1,
            "gpu_per_process": 0,
        },
        "tensorproxy": {
            "enabled": True,
        },
        "user_config": {
            "use_ocdbt": False,
            "use_zarr3": False,
            "checkpoint_path": yt_path,
        },
    }
    with run_config_file(run_config) as run_config_path:
        tracto_cli = TractoCli(
            command=["python3", "/tractorun_tests/tensorproxy_script.py"],
            docker_image=TENSORPROXY_DOCKER_IMAGE,
            args=[
                "--run-config-path",
                run_config_path,
                "--mesh.gpu-per-process",
                "0",
                "--resources.memory-limit",
                "428000000000",
                "--yt-path",
                yt_path,
                "--user-config",
                json.dumps({"use_ocdbt": False, "use_zarr3": False, "checkpoint_path": yt_path}),
                "--bind-local",
                f"{get_data_path('../data/tensorproxy_script.py')}:/tractorun_tests/tensorproxy_script.py",
            ],
        )
        op_run = tracto_cli.run()
    assert op_run.is_exitcode_valid()
    assert op_run.is_operation_state_valid(yt_client=yt_client, job_count=1)



import argparse
import time
from typing import Callable

import yt.wrapper as yt
from yt.wrapper.errors import YtResolveError

from tractorun.private.stderr_reader import (
    YtStderrReader,
    get_job_stderr_with_retry,
)
from tractorun.private.training_dir import TrainingDir


def main() -> None:
    parser = argparse.ArgumentParser(description="Get the stderr of a tractorun peer")
    parser.add_argument(
        "training_dir", metavar="training-dir", type=str, help="The path to the training root directory"
    )
    parser.add_argument(
        "--incarnation",
        type=int,
        help="Number of the incarnation to read stderr. If not set, last incarnation will be used.",
    )
    parser.add_argument(
        "--peer-index",
        type=int,
        default=0,
        help="Index of the peer to read stderr. If not set, primary peer with index 0 will be used.",
    )
    parser.add_argument("-f", "--follow", action="store_true", help="Follow the stderr output")
    args = parser.parse_args()

    training_dir = TrainingDir.create(args.training_dir)
    yt_client = yt.YtClient(config=yt.default_config.get_config_from_env())

    incarnation = args.incarnation
    if incarnation is None:
        try:
            incarnation = yt_client.get(training_dir.base_path + "/@incarnation_id")
        except YtResolveError as e:
            raise Exception("Training dir does not have @incarnation_id attr") from e
        assert incarnation != -1, "No incarnations found"

    incarnation_path = training_dir.get_incarnation_path(incarnation_id=incarnation)
    try:
        yt_client.get(incarnation_path)
    except YtResolveError as e:
        raise Exception(f"Incarnation path is present on cluster {incarnation_path}") from e

    operation_id = yt_client.get(incarnation_path + "/@incarnation_operation_id")
    job_id = yt_client.get(incarnation_path + f"/@topology/{args.peer_index}/job_id")

    stderr_getter = get_job_stderr_with_retry(yt_client=yt_client, operation_id=operation_id, job_id=job_id)

    if args.follow:
        reader = YtStderrReader(stderr_getter=stderr_getter)
        try:
            for line in reader:
                print(line.decode("unicode_escape"), end="")
                time.sleep(1.0)
        except Exception as e:
            print()
            print(e)
    else:
        print_output(stderr_getter)


def print_output(generator: Callable[[], bytes]) -> None:
    stderr = generator().decode("unicode_escape")
    print(stderr)



from typing import Any

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import torch.utils.data

from tractorun.backend.tractorch import (
    Tractorch,
    YtTensorDataset,
)
from tractorun.run import prepare_and_get_toolbox


class Net(nn.Module):
    def __init__(self) -> None:
        super(Net, self).__init__()
        self.l1 = nn.Linear(28 * 28, 10)

    def forward(self, x: Any) -> torch.Tensor:
        return torch.relu(self.l1(x.view(x.size(0), -1)))


if __name__ == "__main__":
    toolbox = prepare_and_get_toolbox(backend=Tractorch())
    user_config = toolbox.get_user_config()
    mnist_ds_path = user_config["MNIST_DS_PATH"]
    device = torch.device("cpu")
    train_dataset = YtTensorDataset(toolbox.yt_client, mnist_ds_path)
    train_loader = torch.utils.data.DataLoader(train_dataset, batch_size=64)
    model = Net().to(device)
    optimizer = optim.Adadelta(model.parameters(), lr=1.0)
    model.train()
    for batch_idx, (data, target) in enumerate(train_loader):
        data, target = data.to(device), target.to(device)
        optimizer.zero_grad()
        output = model(data)
        loss = F.nll_loss(output, target)
        loss.backward()
        optimizer.step()



import sys
import time

import yt.wrapper as yt


def main() -> None:
    path = sys.argv[1]
    value = None
    while value is None:
        value = yt.get(path)
        print(f"sidecar reads {value}", file=sys.stderr)
        yt.set(path, value + 1)
        time.sleep(0.5)


if __name__ == "__main__":
    main()



import argparse
import os
from pathlib import Path
import random
import string

import torch
import torch.distributed as dist

from tractorun.backend.tractorch import Tractorch
from tractorun.mesh import Mesh
from tractorun.resources import Resources
from tractorun.run import run
from tractorun.toolbox import Toolbox


def train(toolbox: Toolbox) -> None:
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    print("Running on device:", device)

    if toolbox.coordinator.get_self_index() == 0:
        x = torch.as_tensor([2.0], device=device)
        y = torch.as_tensor([3.0], device=device)
        print(f"Sending summands to peer 1 (X = {x}, Y = {y})")
        dist.send(x, dst=1)
        dist.send(y, dst=1)
        print("Waiting for result from peer 1")
        sum = torch.zeros(1, device=device)
        dist.recv(tensor=sum, src=1)
        print(f"Received sum from peer 1: {sum}")
    else:
        x = torch.zeros(1, device=device)
        print("Waiting for X from peer 0")
        dist.recv(tensor=x, src=0)
        print(f"Received X from peer 0: {x}")
        print("Waiting for Y from peer 0")
        y = torch.zeros(1, device=device)
        dist.recv(tensor=y, src=0)
        print(f"Received Y from peer 1: {y}")
        sum = x + y
        print(f"Sending sum to peer 1: {sum}")
        dist.send(tensor=sum, dst=0)


def _default_home_dir() -> str:
    rnm = "".join(random.choices(string.ascii_letters + string.digits, k=8))
    return f"//tmp/tractorun_examples/{rnm}"


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--yt-home-dir", type=str, default=_default_home_dir())
    parser.add_argument("--pool-tree", type=str, default="default")
    parser.add_argument("--dataset-path", type=str, default="//home/samples/mnist-torch-train")
    parser.add_argument("--docker-image", type=str, default=os.environ.get("DOCKER_IMAGE"))
    parser.add_argument("--gpu-per-process", type=int, default=0)
    args = parser.parse_args()

    mesh = Mesh(node_count=1, process_per_node=2, gpu_per_process=args.gpu_per_process, pool_trees=[args.pool_tree])

    tractorun_path = (Path(__file__).parent.parent.parent.parent / "tractorun").resolve()
    run(
        train,
        backend=Tractorch(),
        yt_path=args.yt_home_dir,
        mesh=mesh,
        resources=Resources(
            memory_limit=8076021002,
        ),
        docker_image=args.docker_image,
        binds_local_lib=[str(tractorun_path)],
    )



import sys

from tractorun.backend.generic import GenericBackend
from tractorun.run import prepare_and_get_toolbox


def main() -> None:
    toolbox = prepare_and_get_toolbox(backend=GenericBackend())
    self_index = toolbox.coordinator.get_self_index()
    inc = toolbox.coordinator.get_incarnation_id()
    print(f"first stdout line {self_index} {inc}", file=sys.stdout)
    print(f"first stderr line {self_index} {inc}", file=sys.stderr)
    print(f"second stdout line {self_index} {inc}", file=sys.stdout)
    print(f"second stderr line {self_index} {inc}", file=sys.stderr)


if __name__ == "__main__":
    main()



import base64
import os
from pathlib import Path
import pickle
import sys
import time
from typing import Optional
import warnings

import attrs
import yt.wrapper as yt

from tractorun import __version__
from tractorun.env import EnvVariable
from tractorun.exception import TractorunBootstrapError
from tractorun.mesh import Mesh
from tractorun.operation_log import OperationLogMode
from tractorun.private.operation_log import (
    LogHandlerFactory,
    YTLogHandlerFactory,
)
from tractorun.private.process_manager import (
    ProcessManager,
    ProcessManagerPollStatus,
)
from tractorun.private.tensorproxy import TensorproxyBootstrap
from tractorun.private.training_dir import TrainingDir
from tractorun.private.yt_cluster import TractorunClusterConfig
from tractorun.sidecar import Sidecar


PROCESSES_POLL_TIMEOUT = 5


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class LibVersions:
    tractorun: str
    ytsaurus_client: str

    @staticmethod
    def create() -> "LibVersions":
        return LibVersions(
            tractorun=__version__,
            ytsaurus_client=yt.__version__,
        )


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class BootstrapConfig:
    mesh: Mesh
    sidecars: list[Sidecar]
    env: list[EnvVariable]
    training_dir: TrainingDir
    yt_client_config: str
    tensorproxy: Optional[TensorproxyBootstrap]
    lib_versions: LibVersions
    cluster_config: TractorunClusterConfig
    operation_log_mode: OperationLogMode


def check_lib_versions(local_lib_versions: LibVersions) -> None:
    remote_lib_versions = LibVersions.create()
    diff = {}
    if local_lib_versions.tractorun != remote_lib_versions.tractorun:
        diff["tractorun"] = {
            "local": local_lib_versions.tractorun,
            "remote": remote_lib_versions.tractorun,
        }
    if local_lib_versions.ytsaurus_client != remote_lib_versions.ytsaurus_client:
        diff["ytsaurus-client"] = {
            "local": local_lib_versions.ytsaurus_client,
            "remote": remote_lib_versions.ytsaurus_client,
        }
    if diff:
        warnings.warn(f"Local and remote libraries has different versions: {diff}")


def bootstrap(
    mesh: Mesh,
    training_dir: TrainingDir,
    yt_client_config: str,
    command: list[str],
    sidecars: list[Sidecar],
    env: list[EnvVariable],
    tensorproxy: Optional[TensorproxyBootstrap],
    lib_versions: LibVersions,
    cluster_config: TractorunClusterConfig,
    operation_log_mode: OperationLogMode,
    sandbox_path: Path,
) -> None:
    # Runs inside a job

    check_lib_versions(local_lib_versions=lib_versions)

    yt_config = pickle.loads(base64.b64decode(yt_client_config))
    yt_client = yt.YtClient(config=yt_config)

    tp_sidecars, tp_env = [], {}
    if tensorproxy is not None:
        total_ports = mesh.process_per_node + tensorproxy.ports_count
        tp_grpc_port, tp_mon_port = range(total_ports)[mesh.process_per_node :]

        tp_sidecars = tensorproxy.prepare_and_get_sidecars(
            yt_proxy=yt_config["proxy"]["url"],
            grpc_port=tp_grpc_port,
            monitoring_port=tp_mon_port,
            sandbox_path=sandbox_path,
        )
        tp_env = tensorproxy.get_environment(grpc_port=tp_grpc_port)

    spec_env = {}
    for var in env:
        if var.cypress_path is not None:
            spec_env[var.name] = yt_client.get(var.cypress_path)
        else:
            spec_env[var.name] = var.value

    sidecars = sidecars + tp_sidecars

    log_handlers: list[LogHandlerFactory] = []
    if operation_log_mode == OperationLogMode.realtime_yt_table:
        log_handlers.append(
            YTLogHandlerFactory(yt_client_config=yt_client_config, training_dir=training_dir),
        )

    status = ProcessManagerPollStatus.running
    with ProcessManager.start(
        command=command,
        sidecars=sidecars,
        training_dir=training_dir,
        yt_client_config=yt_client_config,
        cluster_config=cluster_config,
        mesh=mesh,
        node_index=int(os.environ["YT_JOB_COOKIE"]),
        os_environ=os.environ,
        tp_env=tp_env,
        spec_env=spec_env,
        log_handler_factories=log_handlers,
        sandbox_path=sandbox_path,
    ) as process_manager:
        while status == ProcessManagerPollStatus.running:
            time.sleep(PROCESSES_POLL_TIMEOUT)
            status = process_manager.poll()
    match status:
        case ProcessManagerPollStatus.success:
            return
        case ProcessManagerPollStatus.fail:
            sys.exit(1)
        case _:
            raise TractorunBootstrapError(f"Unknown poll status {status}")



import json
import sys
from typing import Callable

import pytest

from tests.utils import (
    TRACTORAX_DOCKER_IMAGE,
    TractoCli,
    get_data_path,
)
from tests.yt_instances import YtInstance
from tractorun.mesh import Mesh
from tractorun.run import run
from tractorun.toolbox import Toolbox


def _get_simple_train(mnist_ds_path: str) -> Callable:
    def _simple_train(toolbox: Toolbox) -> None:
        pass

    return _simple_train


def test_run_pickle(can_test_jax: bool, yt_instance: YtInstance, yt_path: str, mnist_ds_path: str) -> None:
    if not can_test_jax:
        pytest.skip("jax can't be run on this platform")

    from jax import (
        grad,
        jit,
    )

    from tractorun.backend.tractorax import Tractorax

    def checker(_: Toolbox) -> None:
        @jit
        def f(x: int) -> int:
            return x**2 + 3 * x + 1

        grad_f = grad(f)
        print(grad_f(1.0), file=sys.stderr)

    yt_client = yt_instance.get_client()

    train_func = _get_simple_train(mnist_ds_path)

    mesh = Mesh(node_count=1, process_per_node=1, gpu_per_process=0)
    # The operation did not fail => success!
    run(
        train_func,
        backend=Tractorax(),
        yt_path=yt_path,
        mesh=mesh,
        yt_client=yt_client,
        docker_image=TRACTORAX_DOCKER_IMAGE,
    )


def test_run_script(yt_instance: YtInstance, yt_path: str, mnist_ds_path: str) -> None:
    yt_client = yt_instance.get_client()

    tracto_cli = TractoCli(
        command=["python3", "/tractorun_tests/jax_run_script.py"],
        docker_image=TRACTORAX_DOCKER_IMAGE,
        args=[
            "--mesh.node-count",
            "1",
            "--mesh.process-per-node",
            "1",
            "--mesh.gpu-per-process",
            "0",
            "--yt-path",
            yt_path,
            "--user-config",
            json.dumps({"MNIST_DS_PATH": mnist_ds_path}),
            "--bind-local",
            f"{get_data_path('../data/jax_run_script.py')}:/tractorun_tests/jax_run_script.py",
        ],
    )
    op_run = tracto_cli.run()
    assert op_run.is_exitcode_valid()
    assert op_run.is_operation_state_valid(yt_client=yt_client, job_count=1)



from tractorun.backend.tractorch.backend import Tractorch
from tractorun.backend.tractorch.dataset import (
    YtDataset,
    YtTensorDataset,
)
from tractorun.backend.tractorch.serializer import TensorSerializer


__all__ = [
    "YtDataset",
    "YtTensorDataset",
    "TensorSerializer",
    "Tractorch",
]



import json
from pathlib import Path
import subprocess
import sys
from typing import Iterable

import pytest

from tests.utils import (
    EXAMPLES_DOCKER_IMAGE,
    TractoCli,
    get_data_path,
)
from tests.yt_instances import YtInstance


def get_examples_pickle_path(example_type: str) -> Iterable[Path]:
    base_examples_path = Path(get_data_path("../../examples")) / example_type
    for path in base_examples_path.rglob("*.py"):
        assert isinstance(path, Path)
        # filter all examples for tractorun cli
        if (path.parent / "run.sh").exists():
            continue
        yield path


def get_examples_script_path(example_type: str) -> Iterable[Path]:
    base_examples_path = Path(get_data_path("../../examples")) / example_type
    for path in base_examples_path.rglob("run.sh"):
        assert isinstance(path, Path)
        script_path = path.parent / (path.parent.name + ".py")
        yield script_path


def run_example_pickle(path: Path, yt_path: str, dataset_path: str) -> None:
    process = subprocess.Popen(
        [
            "python3",
            path,
            "--yt-home-dir",
            yt_path,
            "--dataset-path",
            dataset_path,
            "--docker-image",
            EXAMPLES_DOCKER_IMAGE,
            # run without gpu
            "--pool-tree",
            "default",
            "--gpu-per-process",
            "0",
        ],
        stdout=sys.stdout,
        stderr=sys.stderr,
    )
    process.wait()
    assert process.returncode == 0


def run_example_script(path: Path, yt_path: str, dataset_path: str) -> None:
    tracto_cli = TractoCli(
        command=["python3", get_data_path(path)],
        docker_image=EXAMPLES_DOCKER_IMAGE,
        args=[
            "--yt-path",
            yt_path,
            "--bind-local",
            f"{get_data_path(str(path))}:{get_data_path(str(path))}",
            "--user-config",
            json.dumps({"dataset_path": dataset_path}),
        ],
    )
    op_run = tracto_cli.run()
    assert op_run.is_exitcode_valid()


@pytest.mark.parametrize("example_path", get_examples_pickle_path(example_type="pytorch"))
def test_pytorch_pickle_examples(yt_instance: YtInstance, yt_path: str, mnist_ds_path: str, example_path: Path) -> None:
    run_example_pickle(
        path=example_path,
        yt_path=yt_path,
        dataset_path=mnist_ds_path,
    )


@pytest.mark.parametrize("example_path", get_examples_pickle_path(example_type="jax"))
def test_jax_pickle_examples(
    can_test_jax: bool, yt_instance: YtInstance, yt_path: str, mnist_ds_path: str, example_path: Path
) -> None:
    if example_path.name == "jax_simple_distributed.py":
        pytest.skip("jax_simple_distributed.py can be run only on GPU backend")
    if not can_test_jax:
        pytest.skip("jax can't be run on this platform")
    run_example_pickle(
        path=example_path,
        yt_path=yt_path,
        dataset_path=mnist_ds_path,
    )


@pytest.mark.parametrize("example_path", get_examples_script_path(example_type="pytorch"))
def test_pytorch_script_examples(yt_instance: YtInstance, yt_path: str, mnist_ds_path: str, example_path: Path) -> None:
    run_example_script(
        path=example_path,
        yt_path=yt_path,
        dataset_path=mnist_ds_path,
    )


@pytest.mark.parametrize("example_path", get_examples_script_path(example_type="jax"))
def test_jax_script_examples(
    can_test_jax: bool, yt_instance: YtInstance, yt_path: str, mnist_ds_path: str, example_path: Path
) -> None:
    if not can_test_jax:
        pytest.skip("jax can't be run on this platform")
    run_example_script(
        path=example_path,
        yt_path=yt_path,
        dataset_path=mnist_ds_path,
    )



import base64
import pickle
import threading
import time
from types import TracebackType
from typing import (
    Callable,
    Generator,
    Optional,
    Type,
)

import attrs
from yt import wrapper as yt
from yt.wrapper.errors import YtError

from tractorun.exception import StderrReaderError
from tractorun.mesh import Mesh
from tractorun.private.training_dir import TrainingDir
from tractorun.stderr_reader import StderrMode


YT_RETRY_INTERVAL = 5
STDERR_READER_THREAD_NAME = "tractorun_stderr_reader"
STDERR_READER_STOP_TIMEOUT = 5


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class YtStderrReader:
    _stderr_getter: Callable[[], Optional[bytes]]
    _stop_on_none: bool = attrs.field(default=False)

    def get_output(self) -> Generator[bytes, None, None]:
        last = b""
        while True:
            current = self._stderr_getter()
            if current is None:
                if self._stop_on_none:
                    return
                current = b""
            new_data = self._get_new_data(last, current)
            yield new_data
            last = current

    def _prefix_function(self, s: bytes) -> list[int]:
        pi = [0] * len(s)
        for i in range(1, len(s)):
            j = pi[i - 1]
            while j > 0 and s[i] != s[j]:
                j = pi[j - 1]
            if s[i] == s[j]:
                j += 1
            pi[i] = j
        return pi

    def _get_new_data(self, s: bytes, t: bytes) -> bytes:
        uuid = b"8a5c221d-c111d561-13440384-186"
        k = t + uuid + s
        pi = self._prefix_function(k)
        max_pi = 0
        for i in range(len(t) + len(uuid), len(k)):
            max_pi = max(max_pi, pi[i])
        return t[max_pi:]

    def __iter__(self) -> Generator[bytes, None, None]:
        return self.get_output()


def get_job_stderr(
    yt_client: yt.YtClient,
    operation_id: str,
    job_id: str,
) -> Callable[[], bytes]:
    def _wrapped() -> bytes:
        return yt_client.get_job_stderr(operation_id=operation_id, job_id=job_id).read()

    return _wrapped


def get_job_stderr_with_retry(
    yt_client: yt.YtClient, operation_id: str, job_id: str, retry_interval: float = YT_RETRY_INTERVAL
) -> Callable[[], bytes]:
    def _wrapped() -> bytes:
        while True:
            try:
                data = get_job_stderr(yt_client=yt_client, operation_id=operation_id, job_id=job_id)()
                return data
            except YtError:
                # TODO: add debug logs
                time.sleep(retry_interval)

    return _wrapped


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class StderrReaderWorker:
    _prev_incarnation_id: int
    _training_dir: TrainingDir
    _yt_client_config_pickled: str
    _mode: StderrMode
    _mesh: Mesh
    _stop: bool = False

    _polling_interval: float = attrs.field(default=1.0)
    _yt_retry_interval: float = attrs.field(default=YT_RETRY_INTERVAL)
    _thread: threading.Thread = attrs.field(init=False, default=None)

    def _start(self) -> None:
        yt_config = pickle.loads(base64.b64decode(self._yt_client_config_pickled))
        yt_client = yt.YtClient(config=yt_config)

        incarnation, topology, operation_id = self._prev_incarnation_id, [], None
        while (
            incarnation == self._prev_incarnation_id or operation_id is None or len(topology) != self._mesh.peer_count
        ):
            if self._stop:
                return
            try:
                incarnation = yt_client.get(self._training_dir.base_path + "/@incarnation_id")
                incarnation_path = self._training_dir.get_incarnation_path(incarnation)
                operation_id = yt_client.get(incarnation_path + "/@incarnation_operation_id")
                topology = yt_client.get(incarnation_path + "/@topology")
            except Exception:
                # TODO: add debug logs
                # it's important to reset all values
                # because if stderr get new incarnation id
                # but job doesn't create new incarnation path ->
                # stderr reader starts read previous job's stderr
                incarnation, topology, operation_id = self._prev_incarnation_id, [], None

        match self._mode:
            case StderrMode.primary:
                job_ids = [topology[0]["job_id"]]
            case _:
                raise StderrReaderError(f"Unknown mode {self._mode}")
        output_streams: list[tuple[str, Generator[bytes, None, None]]] = []
        for job_id in job_ids:
            stderr_getter = get_job_stderr(
                yt_client=yt_client,
                operation_id=operation_id,
                job_id=job_id,
            )
            output_streams.append(
                (
                    job_id,
                    YtStderrReader(stderr_getter=stderr_getter).get_output(),
                ),
            )

        while True:
            if self._stop:
                return
            for job_id, output_stream in output_streams:
                try:
                    data = next(output_stream)
                    if data:
                        print(data.decode("utf-8", errors="replace"), end="")
                except Exception:
                    # TODO: add debug logs
                    pass
            time.sleep(self._polling_interval)

    def stop(self) -> None:
        if self._thread is not None:
            # sometimes we lost last lines of stderr
            # TODO: we should check the readiness of artifacts
            time.sleep(STDERR_READER_STOP_TIMEOUT)
            self._stop = True
            self._thread.join(timeout=self._polling_interval * 3)

    def start(self) -> None:
        assert self._thread is None
        if self._mode == StderrMode.disabled:
            return
        self._thread = threading.Thread(target=self._start, name=STDERR_READER_THREAD_NAME)
        self._thread.start()

    def __enter__(self) -> None:
        self.start()

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self.stop()



import logging
from typing import Any
import warnings

import attrs
from yt import wrapper as yt

from tractorun.docker_auth import (
    DockerAuthData,
    DockerAuthSecret,
)
from tractorun.exception import DockerAuthDataError


LOGGER = logging.getLogger(__name__)


CANONICAL_SECRET_FORMAT = """
{
  secrets: {
    username: {
      value: some_username
    },
    password: {
      value: some_password
    },
  }
}
"""


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class DockerAuthInternal:
    username: str | None
    password: str | None
    auth: str | None

    def to_spec(self) -> dict:
        data = {}
        if self.username:
            data["username"] = self.username
        if self.password:
            data["password"] = self.password
        if self.auth:
            data["auth"] = self.auth
        return data


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class DockerAuthDataExtractor:
    _yt_client: yt.YtClient

    def _parse_raw_auth_data_valid_v1(self, raw_auth_data: Any) -> DockerAuthInternal | None:
        LOGGER.debug("Parsing raw auth data from yt client, v1")
        # just backward compatibility
        if not isinstance(raw_auth_data, dict):
            LOGGER.debug("Skip because auth data is not dict")
            return None
        if set(raw_auth_data.keys()) != {"username", "password"} and set(raw_auth_data.keys()) != {"auth"}:
            LOGGER.debug("Skip because of invalid keys: %s", raw_auth_data.keys())
            return None
        if not all([isinstance(value, str) for value in raw_auth_data.values()]):
            LOGGER.debug("Skip because of invalid values types")
            return None
        return DockerAuthInternal(
            username=raw_auth_data.get("username"),
            password=raw_auth_data.get("password"),
            auth=raw_auth_data.get("auth"),
        )

    def _parse_raw_auth_data_valid_v2(self, raw_auth_data: Any) -> DockerAuthInternal | None:
        LOGGER.debug("Parsing raw auth data from yt client, v2")
        if not isinstance(raw_auth_data, dict):
            LOGGER.debug("Skip because auth data is not dict")
            return None
        secrets = raw_auth_data.get("secrets", {})
        if not isinstance(secrets, dict):
            LOGGER.debug("Skip because there is no 'secrets' key at the top level")
            return None
        if set(secrets.keys()) != {"username", "password"} and set(secrets.keys()) != {"auth"}:
            LOGGER.debug("Skip because of invalid keys: %s", secrets.keys())
            return None
        if not all([isinstance(value, dict) for value in secrets.values()]):
            LOGGER.debug("Skip because of invalid values types")
            return None
        if not all([isinstance(value.get("value"), str) for value in secrets.values()]):
            LOGGER.debug("Skip because of invalid values types")
            return None
        return DockerAuthInternal(
            username=secrets.get("username", {}).get("value"),
            password=secrets.get("password", {}).get("value"),
            auth=secrets.get("auth", {}).get("value"),
        )

    def extract(self, auth_data: DockerAuthData) -> DockerAuthInternal:
        match auth_data:
            case DockerAuthSecret():
                assert isinstance(auth_data, DockerAuthSecret)  # skip type warning in pycharm
                LOGGER.debug("Parsing raw auth data from yt client %s", auth_data.cypress_path)
                raw_auth_data = self._yt_client.get(auth_data.cypress_path)
                if parsed_auth_data := self._parse_raw_auth_data_valid_v2(raw_auth_data):
                    return parsed_auth_data
                elif parsed_auth_data := self._parse_raw_auth_data_valid_v1(raw_auth_data):
                    warnings.warn(f"Please use new docker secret format {CANONICAL_SECRET_FORMAT}")
                    return parsed_auth_data
                # TODO: describe how to create new secret in UI
                raise DockerAuthDataError(f"Unknown secret format, please use {CANONICAL_SECRET_FORMAT}")
            case _:
                raise DockerAuthDataError("Unknown auth data type")



YT_USER_CONFIG_ENV_VAR: str = "YT_USER_CONFIG"
TRACTO_CONFIG_ENV_VAR: str = "TRACTO_CONFIG"
BIND_PATHS_ENV_VAR: str = "BIND_PATHS"
BOOTSTRAP_CONFIG_FILENAME_ENV_VAR: str = "BOOTSTRAP_CONFIG_FILENAME"
DEFAULT_CLUSTER_CONFIG_PATH: str = "//sys/tractorun/config"
DEFAULT_TENSORPROXY_PATH = "//sys/tractorun/tensorproxy"
JOB_SANDBOX_PATH: str = "/slot/sandbox"

BOOTSTRAP_CONFIG_NAME: str = "__bootstrap_config"

USER_DESCRIPTION_MANAGER_NAME = "extra"
TRACTORUN_DESCRIPTION_MANAGER_NAME = "tractorun"



from typing import Union

import attrs
from yt import wrapper as yt


__all__ = ["Link", "DescriptionManager", "Description"]

from tractorun.private.yt_cluster import make_cypress_link as _make_cypress_link


@attrs.define(kw_only=True, slots=True, auto_attribs=True, frozen=True)
class Link:
    _value: str | None

    def to_yson(self) -> yt.yson.yson_types.YsonUnicode | None:
        if self._value is None:
            return None
        value = yt.yson.yson_types.YsonUnicode(self._value)
        value.attributes = {"_type_tag": "url"}
        return value


_PRIMITIVE_DESCRIPTION = str | bool | bytes | int | float | Link | None
_COMPLEX_DESCRIPTION = (
    dict[_PRIMITIVE_DESCRIPTION, Union[_PRIMITIVE_DESCRIPTION, "_COMPLEX_DESCRIPTION"]]
    | list[Union[_PRIMITIVE_DESCRIPTION, "_COMPLEX_DESCRIPTION"]]
)
_DESCRIPTION = _COMPLEX_DESCRIPTION | _PRIMITIVE_DESCRIPTION
Description = dict[str, _DESCRIPTION]


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class DescriptionManager:
    _operation_id: str
    _yt_client: yt.YtClient
    _cypress_link_template: str | None
    _key: list[str] = attrs.field(default=[])

    def get_child(self, key: str) -> "DescriptionManager":
        return DescriptionManager(
            operation_id=self._operation_id,
            yt_client=self._yt_client,
            key=self._key + [key],
            cypress_link_template=self._cypress_link_template,
        )

    def make_cypress_link(self, value: str) -> Link | None:
        raw_link = _make_cypress_link(
            cypress_link_template=self._cypress_link_template,
            path=value,
        )
        if raw_link is None:
            return None
        return Link(value=raw_link)

    @classmethod
    def _convert_yson(cls, description: _DESCRIPTION) -> _DESCRIPTION:
        match description:
            case dict():
                return {k: cls._convert_yson(v) for k, v in description.items()}
            case list():
                return [cls._convert_yson(v) for v in description]
            case Link():
                return description.to_yson()
        return description

    @classmethod
    def _make_description(cls, key: list[str], description: Description) -> _DESCRIPTION:
        result_description: dict = {}
        current_part = result_description
        for k in key:
            current_part[k] = {}
            current_part = current_part[k]
        current_part.update(description)
        converted_description = cls._convert_yson(result_description)
        return converted_description

    def set(self, description: Description) -> None:
        new_description = self._make_description(self._key, description)
        self._yt_client.update_operation_parameters(
            self._operation_id,
            parameters={
                "annotations": {"description": new_description},
            },
        )



import abc

import attrs


__all__ = ["DockerAuthSecret", "DockerAuthData"]


class DockerAuthData(abc.ABC):
    pass


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class DockerAuthSecret(DockerAuthData):
    cypress_path: str



import argparse
import os
from pathlib import Path
import random
import string
import sys
from typing import (
    Any,
    Tuple,
)

from pytorch_lightning import (
    LightningModule,
    Trainer,
)
from pytorch_lightning.utilities.types import STEP_OUTPUT
import torch
from torch import Tensor
from torch.nn import functional as F
from torch.utils.data import DataLoader

from tractorun.backend.tractorch import Tractorch
from tractorun.backend.tractorch.dataset import YtTensorDataset
from tractorun.mesh import Mesh
from tractorun.resources import Resources
from tractorun.run import run
from tractorun.toolbox import Toolbox


class MNISTModel(LightningModule):
    def __init__(self) -> None:
        super().__init__()
        self.l1 = torch.nn.Linear(28 * 28, 10)

    def forward(self, x: Any) -> torch.Tensor:
        return torch.relu(self.l1(x.view(x.size(0), -1)))

    def training_step(self, batch: Tuple[Tensor, ...], batch_nb: Any) -> STEP_OUTPUT:
        x, y = batch
        loss = F.cross_entropy(self(x), y)
        return loss

    def configure_optimizers(self) -> torch.optim.Adam:
        return torch.optim.Adam(self.parameters(), lr=0.02)


def train(toolbox: Toolbox) -> None:
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    print("Running on device:", device, file=sys.stderr)

    mnist_model = MNISTModel()
    dataset_path = toolbox.get_user_config()["dataset_path"]
    train_dataset = YtTensorDataset(
        toolbox.yt_client,
        dataset_path,
        columns=["data", "labels"],
        start=0,
        end=20,
    )
    train_loader = DataLoader(train_dataset, batch_size=64)

    trainer = Trainer(max_epochs=1)

    trainer.fit(mnist_model, train_loader)


def _default_home_dir() -> str:
    rnm = "".join(random.choices(string.ascii_letters + string.digits, k=8))
    return f"//tmp/tractorun_examples/{rnm}"


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--yt-home-dir", type=str, default=_default_home_dir())
    parser.add_argument("--pool-tree", type=str, default="default")
    parser.add_argument("--dataset-path", type=str, default="//home/samples/mnist-torch-train")
    parser.add_argument("--docker-image", type=str, default=os.environ.get("DOCKER_IMAGE"))
    parser.add_argument("--gpu-per-process", type=int, default=0)
    args = parser.parse_args()

    mesh = Mesh(node_count=1, process_per_node=8, gpu_per_process=args.gpu_per_process, pool_trees=[args.pool_tree])

    tractorun_path = (Path(__file__).parent.parent.parent.parent / "tractorun").resolve()
    run(
        train,
        backend=Tractorch(),
        yt_path=args.yt_home_dir,
        mesh=mesh,
        docker_image=args.docker_image,
        resources=Resources(
            memory_limit=8076021002,
        ),
        user_config={
            "dataset_path": args.dataset_path,
        },
        binds_local_lib=[str(tractorun_path)],
    )



import attrs

from tractorun.backend.generic.environment import Environment
from tractorun.base_backend import BackendBase


__all__ = ["GenericBackend"]


@attrs.define
class GenericBackend(BackendBase):
    @property
    def environment(self) -> Environment:
        return Environment()



import attrs

from tractorun.description import Link
from tractorun.mesh import Mesh


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class TractorunDescription:
    training_dir: Link
    primary_stderr: Link
    logs: Link
    primary_address: str
    incarnation: int
    mesh: Mesh

    def to_dict(self) -> dict:
        return {
            "training_dir": self.training_dir,
            "primary": {
                "job_stderr": self.primary_stderr,
                "address": self.primary_address,
            },
            "logs": self.logs,
            "incarnation": self.incarnation,
            "mesh": attrs.asdict(self.mesh),  # type: ignore
        }



#!/usr/bin/env python3

import argparse
import json
import sys
import traceback
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    cast,
    overload,
)

from attr import AttrsInstance
import attrs
from cattrs import ClassValidationError
import yaml

from tractorun.bind import (
    BindAttributes,
    BindCypress,
    BindLocal,
)
from tractorun.docker_auth import (
    DockerAuthData,
    DockerAuthSecret,
)
from tractorun.env import EnvVariable
from tractorun.exception import TractorunConfigError
from tractorun.mesh import Mesh
from tractorun.operation_log import OperationLogMode
from tractorun.private.constants import (
    DEFAULT_CLUSTER_CONFIG_PATH,
    DEFAULT_TENSORPROXY_PATH,
)
from tractorun.private.docker_auth import DockerAuthInternal
from tractorun.private.helpers import (
    create_attrs_converter,
    get_default_docker_image,
)
from tractorun.resources import Resources
from tractorun.run import run_script
from tractorun.run_info import RunInfo
from tractorun.sidecar import (
    RestartPolicy,
    Sidecar,
)
from tractorun.stderr_reader import StderrMode
from tractorun.tensorproxy import TensorproxySidecar


__default_mesh = Mesh()

MESH_NODE_COUNT_DEFAULT = __default_mesh.node_count
MESH_PROCESS_PER_NODE_DEFAULT = __default_mesh.process_per_node
MESH_GPU_PER_PROCESS_DEFAULT = __default_mesh.gpu_per_process
TENSORPROXY_ENABLED_DEFAULT = False
CLUSTER_CONFIG_PATH_DEFAULT = DEFAULT_CLUSTER_CONFIG_PATH
TENSORPROXY_RESTART_POLICY_DEFAULT = RestartPolicy.ALWAYS
LOCAL_DEFAULT = False
NO_WAIT_DEFAULT = False
PROXY_STDERR_MODE_DEFAULT = StderrMode.disabled
OPERATION_LOG_MODE_DEFAULT = OperationLogMode.default


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class MeshConfig:
    node_count: int | None = attrs.field(default=None)
    process_per_node: int | None = attrs.field(default=None)
    gpu_per_process: int | None = attrs.field(default=None)
    pool_trees: list[str] | None = attrs.field(default=None)
    pool: str | None = attrs.field(default=None)


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class ResourcesConfig:
    cpu_limit: float | None = attrs.field(default=None)
    memory_limit: int | None = attrs.field(default=None)


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class TensorproxyConfig:
    enabled: bool | None = attrs.field(default=None)
    restart_policy: RestartPolicy | None = attrs.field(default=None)
    yt_path: str | None = attrs.field(default=None)


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class DockerAuthSecretConfig:
    cypress_path: str


_T = TypeVar("_T")


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class Config:
    """yaml config representation"""

    yt_path: str | None = attrs.field(default=None)
    docker_image: str | None = attrs.field(default=None)
    title: str | None = attrs.field(default=None)
    user_config: dict[str, Any] | None = attrs.field(default=None)
    yt_operation_spec: dict[str, Any] | None = attrs.field(default=None)
    yt_task_spec: dict[str, Any] | None = attrs.field(default=None)
    local: bool | None = attrs.field(default=None)
    no_wait: bool | None = attrs.field(default=None)
    bind_local: list[BindLocal] | None = attrs.field(default=None)
    bind_local_lib: list[str] | None = attrs.field(default=None)
    bind_cypress: list[BindCypress] | None = attrs.field(default=None)
    proxy_stderr_mode: StderrMode | None = attrs.field(default=None)
    operation_log_mode: OperationLogMode | None = attrs.field(default=None)
    cluster_config_path: str | None = attrs.field(default=None)
    command: list[str] | None = attrs.field(default=None)

    mesh: MeshConfig = attrs.field(default=MeshConfig())
    resources: ResourcesConfig = attrs.field(default=ResourcesConfig())
    sidecars: list[Sidecar] | None = attrs.field(default=None)
    env: list[EnvVariable] | None = attrs.field(default=None)
    tensorproxy: TensorproxyConfig = attrs.field(default=TensorproxyConfig())
    docker_auth_secret: DockerAuthSecretConfig | None = attrs.field(default=None)

    @classmethod
    def load_yaml(cls, path: str) -> "Config":
        with open(path, "r") as f:
            config = yaml.safe_load(f)
        converter = create_attrs_converter()
        try:
            return converter.structure(config, Config)
        except ClassValidationError as e:
            raise TractorunConfigError from e


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class EffectiveConfig:
    yt_path: str
    docker_image: str
    title: str | None
    user_config: dict[str, Any]
    yt_operation_spec: dict[str, Any]
    yt_task_spec: dict[str, Any]
    local: bool
    no_wait: bool
    bind_local: list[BindLocal]
    bind_local_lib: list[str]
    bind_cypress: list[BindCypress]
    proxy_stderr_mode: StderrMode
    operation_log_mode: OperationLogMode
    cluster_config_path: str
    command: list[str]

    mesh: Mesh
    resources: Resources
    sidecars: list[Sidecar]
    env: list[EnvVariable]
    tensorproxy: TensorproxySidecar
    docker_auth_secret: DockerAuthData | None
    dry_run: bool

    @classmethod
    def configure(cls, args: dict[str, Any], config: Config) -> "EffectiveConfig":
        @overload
        def _choose_value(args_value: _T | None, config_value: _T | None, default: None = None) -> _T | None: ...

        @overload
        def _choose_value(args_value: _T | None, config_value: _T | None, default: _T) -> _T: ...

        def _choose_value(args_value: _T | None, config_value: _T | None, default: _T | None = None) -> _T | None:
            result = args_value if args_value is not None else config_value
            if result is None and default is not None:
                return default
            return result

        command = _choose_value(args["command"] or None, config.command)
        if not command:
            raise TractorunConfigError("Command should be set in config or by cli param")
        yt_path = _choose_value(args["yt_path"], config.yt_path)
        if yt_path is None:
            raise TractorunConfigError("Command should be set in config or by cli param --yt-path")
        docker_image = _choose_value(
            args_value=args["docker_image"],
            config_value=config.docker_image,
            default=get_default_docker_image(),
        )
        if docker_image is None:
            raise TractorunConfigError("docker_image should be set in config or by cli param")

        new_config = EffectiveConfig(
            yt_path=yt_path,
            docker_image=docker_image,
            title=_choose_value(args_value=args["title"], config_value=config.title),
            user_config=_choose_value(args_value=args["user_config"], config_value=config.user_config, default={}),
            yt_operation_spec=_choose_value(
                args_value=args["yt_operation_spec"],
                config_value=config.yt_operation_spec,
                default={},
            ),
            yt_task_spec=_choose_value(args_value=args["yt_task_spec"], config_value=config.yt_task_spec, default={}),
            local=_choose_value(args_value=args["local"], config_value=config.local, default=LOCAL_DEFAULT),
            no_wait=_choose_value(args_value=args["no_wait"], config_value=config.no_wait, default=NO_WAIT_DEFAULT),
            cluster_config_path=_choose_value(
                args_value=args["cluster_config_path"],
                config_value=config.cluster_config_path,
                default=CLUSTER_CONFIG_PATH_DEFAULT,
            ),
            bind_local=_choose_value(
                args_value=args["bind_local"],
                config_value=config.bind_local,
                default=[],
            ),
            bind_local_lib=_choose_value(
                args_value=args["bind_local_lib"], config_value=config.bind_local_lib, default=[]
            ),
            bind_cypress=_choose_value(
                args_value=args["bind_cypress"],
                config_value=config.bind_cypress,
                default=[],
            ),
            proxy_stderr_mode=_choose_value(
                args_value=args["proxy_stderr_mode"],
                config_value=config.proxy_stderr_mode,
                default=PROXY_STDERR_MODE_DEFAULT,
            ),
            operation_log_mode=_choose_value(
                args_value=args["operation_log_mode"],
                config_value=config.operation_log_mode,
                default=OPERATION_LOG_MODE_DEFAULT,
            ),
            sidecars=_choose_value(args_value=args["sidecar"], config_value=config.sidecars, default=[]),
            env=_choose_value(args_value=args["env"], config_value=config.env, default=[]),
            command=command,
            mesh=Mesh(
                node_count=_choose_value(
                    args_value=args["mesh.node_count"],
                    config_value=config.mesh.node_count,
                    default=MESH_NODE_COUNT_DEFAULT,
                ),
                process_per_node=_choose_value(
                    args_value=args["mesh.process_per_node"],
                    config_value=config.mesh.process_per_node,
                    default=MESH_PROCESS_PER_NODE_DEFAULT,
                ),
                gpu_per_process=_choose_value(
                    args_value=args["mesh.gpu_per_process"],
                    config_value=config.mesh.gpu_per_process,
                    default=MESH_GPU_PER_PROCESS_DEFAULT,
                ),
                pool_trees=_choose_value(
                    args_value=args["mesh.pool_trees"],
                    config_value=config.mesh.pool_trees,
                ),
                pool=_choose_value(
                    args_value=args["mesh.pool"],
                    config_value=config.mesh.pool,
                ),
            ),
            resources=Resources(
                cpu_limit=_choose_value(
                    args_value=args["resources.cpu_limit"],
                    config_value=config.resources.cpu_limit,
                ),
                memory_limit=_choose_value(
                    args_value=args["resources.memory_limit"],
                    config_value=config.resources.memory_limit,
                ),
            ),
            tensorproxy=TensorproxySidecar(
                enabled=_choose_value(
                    args_value=args["tensorproxy.enabled"],
                    config_value=config.tensorproxy.enabled,
                    default=TENSORPROXY_ENABLED_DEFAULT,
                ),
                yt_path=_choose_value(
                    args_value=args["tensorproxy.yt_path"],
                    config_value=config.tensorproxy.yt_path,
                    default=DEFAULT_TENSORPROXY_PATH,
                ),
                restart_policy=_choose_value(
                    args_value=args["tensorproxy.restart_policy"],
                    config_value=config.tensorproxy.restart_policy,
                    default=TENSORPROXY_RESTART_POLICY_DEFAULT,
                ),
            ),
            docker_auth_secret=cls._make_docker_auth_secret(
                _choose_value(
                    args_value=args["docker_auth_secret.cypress_path"],
                    # TODO: make some helper
                    config_value=getattr(config.docker_auth_secret, "cypress_path", None),
                ),
            ),
            dry_run=args["dry_run"],
        )
        return new_config

    @classmethod
    def _make_docker_auth_secret(cls, docker_auth_cypress_path: str | None) -> DockerAuthSecret | None:
        if docker_auth_cypress_path is None:
            return None
        return DockerAuthSecret(cypress_path=docker_auth_cypress_path)


def _load_json(value: str | None) -> dict | None:
    if value is None:
        return None
    return json.loads(value)


def _parse_bind_local_arg(value: str | None) -> BindLocal | None:
    if value is None:
        return None
    if value.startswith("{"):
        return BindLocal(**json.loads(value))
    source, destination = value.split(":")
    return BindLocal(source=source, destination=destination)


def _parse_bind_cypress_arg(value: str | None) -> BindCypress | None:
    if value is None:
        return None
    if value.startswith("{"):
        converter = create_attrs_converter()
        try:
            return converter.structure(json.loads(value), BindCypress)
        except ClassValidationError as e:
            raise TractorunConfigError from e
    source, destination = value.split(":")
    return BindCypress(source=source, destination=destination)


def _attr_to_json(value: object) -> str:
    if TYPE_CHECKING:
        value = cast(AttrsInstance, value)
    return json.dumps(attrs.asdict(value))


def make_cli_parser() -> argparse.ArgumentParser:
    # Defaults shouldn't be set in argparse
    # because it interferes with merging with configs.
    # We don't want to use config values as defaults too.
    parser = argparse.ArgumentParser(
        description="Tractorun",
    )
    parser.add_argument("--yt-path", help="YT workdir", type=str)
    parser.add_argument("--run-config-path", type=str, help="path to tractorun config")
    parser.add_argument("--docker-image", type=str, help=f"docker image name. Default: {get_default_docker_image()}")
    parser.add_argument(
        "--docker-auth-secret.cypress-path",
        type=str,
        help="path to the cypress node with format {}".format(
            json.dumps(
                _attr_to_json(DockerAuthInternal(username="placeholder", password="placeholder", auth="placeholder")),  # type: ignore
            ),
        ),
        default=None,
    )
    parser.add_argument("--title", type=str, help="title of the operation")
    parser.add_argument("--mesh.node-count", type=int, help=f"mesh node count. Default: {MESH_NODE_COUNT_DEFAULT}")
    parser.add_argument(
        "--mesh.process-per-node", type=int, help=f"mesh process per node. Default: {MESH_PROCESS_PER_NODE_DEFAULT}"
    )
    parser.add_argument(
        "--mesh.gpu-per-process", type=int, help=f"mesh gpu per process. Default: {MESH_GPU_PER_PROCESS_DEFAULT}"
    )
    parser.add_argument("--mesh.pool-trees", help="mesh pool trees", action="append")
    parser.add_argument("--mesh.pool", help="mesh pool", type=str)
    parser.add_argument("--resources.cpu-limit", type=int, help="cpu limit")
    parser.add_argument("--resources.memory-limit", type=int, help="mem limit")
    parser.add_argument("--user-config", type=_load_json, help="json config that will be passed to the jobs")
    parser.add_argument(
        "--cluster-config-path",
        type=str,
        help=f"path to the global tractorun config on YTSaurus cluster. Default: {CLUSTER_CONFIG_PATH_DEFAULT}",
    )
    parser.add_argument(
        "--tensorproxy.enabled",
        type=bool,
        help=f"enable tensorproxy sidecar. Default {TENSORPROXY_ENABLED_DEFAULT}",
    )
    parser.add_argument(
        "--tensorproxy.yt_path",
        type=str,
        help=f"YTsaurus path to tensorproxy binary. Default {DEFAULT_TENSORPROXY_PATH}",
    )
    parser.add_argument(
        "--tensorproxy.restart_policy",
        type=str,
        help=f"Tensorflow's sidecar restart policy. Default {TENSORPROXY_RESTART_POLICY_DEFAULT}",
    )
    parser.add_argument("--yt-operation-spec", help="yt operation spec", type=_load_json)
    parser.add_argument("--yt-task-spec", help="yt task spec", type=_load_json)
    # TODO: rename to run mode
    parser.add_argument("--local", type=bool, help=f"enable local run mode. Default {LOCAL_DEFAULT}")
    parser.add_argument(
        "--bind-local",
        type=_parse_bind_local_arg,
        action="append",
        default=None,
        help="bind local file or folder to be passed to the docker container. Format: `local_path:remote_path` or {}".format(
            _attr_to_json(
                BindLocal(
                    source="placeholder",
                    destination="placeholder",
                ),
            ),
        ),
    )
    parser.add_argument(
        "--bind-local-lib",
        type=str,
        action="append",
        default=None,
        help="path to local python library to bind it to the remote container and remote PYTHONPATH",
    )
    parser.add_argument(
        "--bind-cypress",
        type=_parse_bind_cypress_arg,
        action="append",
        default=None,
        help="bind cypress file to be passed to the docker container. Format: `local_path:remote_path` or {0}".format(
            _attr_to_json(
                BindCypress(
                    source="yt path",
                    destination="path inside job",
                    attributes=BindAttributes(),
                ),
            ),
        ),
    )
    parser.add_argument(
        "--sidecar",
        action="append",
        type=Sidecar.from_args,
        help="sidecar in json format `{example}`. Restart policy: {policy}".format(
            policy=", ".join(p for p in RestartPolicy),
            example=_attr_to_json(Sidecar(command=["placeholder"], restart_policy=RestartPolicy.ALWAYS)),
        ),
    )
    parser.add_argument(
        "--proxy-stderr-mode",
        type=StderrMode,
        help="proxy jobs stderr to terminal. Mode: "
        + ", ".join(m for m in StderrMode)
        + f". Default: {PROXY_STDERR_MODE_DEFAULT}",
    )
    parser.add_argument(
        "--operation-log-mode",
        type=OperationLogMode,
        help="store operation log mode. Mode: "
        + ", ".join(m for m in OperationLogMode)
        + f". Default: {OPERATION_LOG_MODE_DEFAULT}",
    )
    parser.add_argument(
        "--no-wait",
        action="store_true",
        default=None,
        help=f"don't create transaction and don't wait for operation to complete. Default {NO_WAIT_DEFAULT}",
    )
    parser.add_argument(
        "--env",
        action="append",
        type=EnvVariable.from_args,
        help="set env variable by value or from cypress node. JSON message like {} or {}".format(
            _attr_to_json(
                EnvVariable(
                    name="placeholder",
                    value="placeholder",
                ),
            ),
            _attr_to_json(
                EnvVariable(
                    name="placeholder",
                    cypress_path="placeholder",
                ),
            ),
        ),
    )
    parser.add_argument("--dry-run", help="get internal information without running an operation", action="store_true")
    parser.add_argument("command", nargs="*", help="command to run")
    return parser


def make_configuration(cli_args: list) -> tuple[dict, Config, EffectiveConfig]:
    parser = make_cli_parser()
    args = vars(parser.parse_args(cli_args))
    file_config_content = Config.load_yaml(args["run_config_path"]) if args["run_config_path"] else Config()
    effective_config = EffectiveConfig.configure(
        config=file_config_content,
        args=args,
    )
    return args, file_config_content, effective_config


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class ConfigurationDebug:
    file_config: Config
    effective_config: EffectiveConfig
    cli_args: dict


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class CliRunInfo:
    configuration: ConfigurationDebug
    run_info: RunInfo | None
    errors: list[Any]


def main() -> None:
    args, file_config_content, effective_config = make_configuration(sys.argv[1:])

    run_info = None
    errors = []
    try:
        run_info = run_script(
            user_command=effective_config.command,
            mesh=effective_config.mesh,
            title=effective_config.title,
            resources=effective_config.resources,
            yt_path=effective_config.yt_path,
            docker_image=effective_config.docker_image,
            binds_local=effective_config.bind_local,
            binds_local_lib=effective_config.bind_local_lib,
            binds_cypress=effective_config.bind_cypress,
            tensorproxy=effective_config.tensorproxy,
            proxy_stderr_mode=effective_config.proxy_stderr_mode,
            operation_log_mode=effective_config.operation_log_mode,
            sidecars=effective_config.sidecars,
            env=effective_config.env,
            user_config=effective_config.user_config,
            cluster_config_path=effective_config.cluster_config_path,
            yt_operation_spec=effective_config.yt_operation_spec,
            yt_task_spec=effective_config.yt_task_spec,
            no_wait=effective_config.no_wait,
            docker_auth=effective_config.docker_auth_secret,
            dry_run=effective_config.dry_run,
            yt_client=None,
            local=effective_config.local,
        )
    except Exception:
        errors.append(traceback.format_exc())
        if not effective_config.dry_run:
            raise
    if effective_config.dry_run or effective_config.no_wait:
        cli_run_info = CliRunInfo(
            configuration=ConfigurationDebug(
                file_config=file_config_content,
                cli_args=args,
                effective_config=effective_config,
            ),
            run_info=run_info,
            errors=errors,
        )
        print(
            json.dumps(
                attrs.asdict(cli_run_info),  # type: ignore
                indent=4,
                default=_json_encoder,
            ),
        )


def _json_encoder(value: Any) -> Any:
    if isinstance(value, bytes):
        return value.decode("utf-8")
    return str(value)


if __name__ == "__main__":
    main()



from typing import (
    TYPE_CHECKING,
    Callable,
    Iterator,
    Sized,
    TypeVar,
)
import warnings

import attrs
import torch.utils.data
from yt import wrapper as yt

from tractorun.backend.tractorch.serializer import TensorSerializer
from tractorun.toolbox import Toolbox


T_co = TypeVar("T_co")


__all__ = ["YTTensorTransform", "YtTensorDataset", "YtDataset"]


@attrs.define(frozen=True, slots=True)
class YTTensorTransform:
    _serializer: TensorSerializer = attrs.field(default=TensorSerializer())

    def __call__(self, columns: list[str], row: dict) -> tuple:
        return tuple((self._serializer.desirialize(yt.yson.get_bytes(row[name])) for name in columns))


class YtDataset(torch.utils.data.IterableDataset[T_co], Sized):
    def __init__(
        self,
        path: str,
        yt_client: yt.YtClient,
        transform: Callable[[list[str], dict], T_co],
        start: int = 0,
        end: int | None = None,
        columns: list | None = None,
    ) -> None:
        self._yt_client = yt_client

        row_count = self._yt_client.get(path + "/@row_count")
        if end is None:
            end = row_count
        else:
            assert end <= row_count

        all_columns = []
        for column in self._yt_client.get(path + "/@schema"):
            if TYPE_CHECKING:
                assert isinstance(column["name"], str)
            all_columns.append(column["name"])
        if columns is None:
            columns = all_columns
        else:
            for column in columns:
                assert column in all_columns

        read_path = f"{path}[#{start}:#{end}]"
        self._len = end - start
        self._read_path = read_path
        self._columns = columns
        self._serializer = TensorSerializer()
        self._transform = transform

    def __iter__(self) -> Iterator[T_co]:
        return (self._transform(self._columns, row) for row in self._yt_client.read_table(self._read_path))

    def __len__(self) -> int:
        return self._len


class YtTensorDataset(YtDataset[tuple]):
    def __init__(
        self,
        yt_client: yt.YtClient,
        path: str,
        start: int = 0,
        toolbox: Toolbox | None = None,
        end: int | None = None,
        columns: list | None = None,
    ) -> None:
        if yt_client is not None:
            self._yt_client = yt_client
        # for migration
        elif toolbox is not None:
            warnings.warn("Set yt_client instead of toolbox")
            self._yt_client = toolbox.yt_client
        else:
            raise ValueError("missing 1 required positional argument: yt_client")

        super().__init__(
            yt_client=yt_client,
            path=path,
            start=start,
            end=end,
            columns=columns,
            transform=YTTensorTransform(),
        )



import sys
import time

from yt import wrapper as yt

from tractorun.backend.generic import GenericBackend
from tractorun.run import prepare_and_get_toolbox


def main() -> None:
    toolbox = prepare_and_get_toolbox(backend=GenericBackend())
    user_config = toolbox.get_user_config()

    yt_path = user_config["yt_path"]
    attr_key = user_config["attr_key"]
    attr_value = user_config["attr_value"]

    client = toolbox.yt_client
    value = None
    attempts = 0
    while value is None:
        attr_path = f"{yt_path}/@{attr_key}"
        print(f"try to read value from {attr_path}", file=sys.stderr)
        try:
            value = client.get(attr_path)
        except yt.errors.YtResolveError:
            pass
        time.sleep(5)
        if attempts > 5:
            raise Exception("Something wrong with sidecar")
        attempts += 1
    print("its ok", file=sys.stderr)
    assert value == attr_value


if __name__ == "__main__":
    main()



import sys

import attrs

from tractorun.mesh import Mesh


__all__ = ["Coordinator"]


@attrs.define(kw_only=True)
class Coordinator:
    _mesh: Mesh
    _self_index: int
    _process_index: int
    _self_endpoint: str
    _incarnation_id: int
    _primary_endpoint: str

    def get_self_index(self) -> int:
        return self._self_index

    def get_total_peer_count(self) -> int:
        return self._mesh.peer_count

    def get_incarnation_id(self) -> int:
        return self._incarnation_id

    def is_primary(self) -> bool:
        return self._self_index == 0

    def get_primary_endpoint(self) -> str:
        if self.is_primary():
            port = self._primary_endpoint.split(":")[1]
            print("original primary address {0} was changed to {127.0.0.1}", file=sys.stderr)
            return f"127.0.0.1:{port}"
        return self._primary_endpoint

    def get_process_index(self) -> int:
        return self._process_index



import json
import os
from typing import Any

import attrs
import yt.wrapper as yt

from tractorun.checkpoint import CheckpointManager
from tractorun.coordinator import Coordinator
from tractorun.description import DescriptionManager
from tractorun.mesh import Mesh
from tractorun.private import constants as _constants
from tractorun.private.closet import TrainingMetadata as _TrainingMetadata
from tractorun.private.training_dir import TrainingDir as _TrainingDir


__all__ = ["Toolbox"]


@attrs.define
class Toolbox:
    coordinator: Coordinator
    checkpoint_manager: CheckpointManager
    description_manager: DescriptionManager
    yt_client: yt.YtClient
    mesh: Mesh
    _training_dir: _TrainingDir
    _training_metadata: _TrainingMetadata

    @staticmethod
    def get_user_config() -> dict[Any, Any]:
        return json.loads(os.environ[_constants.YT_USER_CONFIG_ENV_VAR])

    def save_model(self, data: bytes, dataset_path: str | None = None, metadata: dict[str, str] | None = None) -> str:
        incarnation_id = self.coordinator.get_incarnation_id()
        path = self._training_dir.models_path + f"/{incarnation_id}"

        if not self.coordinator.is_primary():
            return path

        self.yt_client.write_file(path, data)
        attributes = {
            "dataset_path": dataset_path,
            "operation_id": self._training_metadata.operation_id,
            "job_id": self._training_metadata.job_id,
            "incarnation_id": incarnation_id,
            "metadata": metadata or {},
            "user_config": self.get_user_config(),
            "mesh": attrs.asdict(self.mesh),  # type: ignore
        }
        self.yt_client.set_attribute(
            path=path,
            attribute="tractorun",
            value=attributes,
        )
        return path



import argparse
import os
from pathlib import Path
import random
import string
import sys
from typing import Any

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import torch.utils.data

from tractorun.backend.tractorch import Tractorch
from tractorun.backend.tractorch.dataset import YtTensorDataset
from tractorun.backend.tractorch.serializer import TensorSerializer
from tractorun.mesh import Mesh
from tractorun.resources import Resources
from tractorun.run import run
from tractorun.toolbox import Toolbox


class Net(nn.Module):
    def __init__(self) -> None:
        super(Net, self).__init__()
        self.l1 = torch.nn.Linear(28 * 28, 10)

    def forward(self, x: Any) -> torch.Tensor:
        return torch.relu(self.l1(x.view(x.size(0), -1)))


def train(toolbox: Toolbox) -> None:
    dataset_path = toolbox.get_user_config()["dataset_path"]
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    serializer = TensorSerializer()
    print("Running on device:", device, file=sys.stderr)

    train_dataset = YtTensorDataset(
        toolbox.yt_client,
        dataset_path,
        columns=["data", "labels"],
        start=0,
        end=10,
    )
    dataset_len = len(train_dataset)

    train_loader = torch.utils.data.DataLoader(train_dataset, batch_size=64)

    model = Net().to(device)
    optimizer = optim.Adadelta(model.parameters(), lr=1.0)
    model.train()
    for batch_idx, (data, target) in enumerate(train_loader):
        data, target = data.to(device), target.to(device)
        optimizer.zero_grad()
        output = model(data)
        loss = F.nll_loss(output, target)
        loss.backward()
        optimizer.step()
        if batch_idx % 10 == 0:
            print(
                "Train[{}/{} ({:.0f}%)]\tLoss: {:.6f}".format(
                    batch_idx * len(data),
                    dataset_len,
                    100.0 * batch_idx / len(train_loader),
                    loss.item(),
                ),
                file=sys.stderr,
            )

    model_path = toolbox.save_model(
        data=serializer.serialize(model.state_dict()),
        dataset_path=dataset_path,
        metadata={},
    )
    print("Model saved to", model_path, file=sys.stderr)


def _default_home_dir() -> str:
    rnm = "".join(random.choices(string.ascii_letters + string.digits, k=8))
    return f"//tmp/tractorun_examples/{rnm}"


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--yt-home-dir", type=str, default=_default_home_dir())
    parser.add_argument("--pool-tree", type=str, default="default")
    parser.add_argument("--dataset-path", type=str, default="//home/samples/mnist-torch-train")
    parser.add_argument("--docker-image", type=str, default=os.environ.get("DOCKER_IMAGE"))
    parser.add_argument("--gpu-per-process", type=int, default=0)
    args = parser.parse_args()

    mesh = Mesh(node_count=1, process_per_node=8, gpu_per_process=args.gpu_per_process, pool_trees=[args.pool_tree])

    tractorun_path = (Path(__file__).parent.parent.parent.parent / "tractorun").resolve()
    run(
        train,
        backend=Tractorch(),
        yt_path=args.yt_home_dir,
        resources=Resources(
            memory_limit=8076021002,
        ),
        mesh=mesh,
        docker_image=args.docker_image,
        binds_local_lib=[str(tractorun_path)],
        user_config={"dataset_path": args.dataset_path},
    )
